package yartiss.engine.simulation;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.TreeSet;

import yartiss.engine.simulation.events.DeadlineMissedEvent;
import yartiss.engine.simulation.events.EnergyFailureEvent;
import yartiss.engine.simulation.events.EventGenerator;
import yartiss.engine.simulation.events.EventVisitorAdapter;
import yartiss.engine.simulation.events.IEventGeneratedListener;
import yartiss.engine.simulation.io.SimulationWriter;
import yartiss.engine.simulation.logger.Logger;
import yartiss.engine.simulation.statistics.StatisticsManager;
import yartiss.engine.simulation.system.ITask;

public class Simulation {

	private SimulationParameters parameters;
	private TreeSet<Event> eventQueue;
	private Scheduler scheduler;
	private final SimulationStatistics statistics = new SimulationStatistics();
	private String logFile = "simulation_1";
	private SimulationWriter simulationWriter = new SimulationWriter();
	private EventGenerator eventGen;


	/**
	 * @return the simulationParameters
	 */
	public SimulationParameters getSimulationParameters() {
		return parameters;
	}

	/**
	 * @param simulationParameters the simulationParameters to set
	 */
	public void setSimulationParameters(SimulationParameters simulationParameters) {
		this.parameters = simulationParameters;
		Processor [] prcs = new Processor [simulationParameters.getNbProc()];
		for (int i = 0; i < prcs.length; i++) {
			prcs[i] = new Processor(i+1, 20,20);
		}
		scheduler.setProcessors(prcs);
		for (IStatisticCriterion criterion : statisticsEventListeners) {
			criterion.setSimulationParameters(parameters);
		}
	}

	/**
	 * 
	 * @param simulationParameters
	 */
	public Simulation(SimulationParameters simulationParameters) {
		this();
		setSimulationParameters(simulationParameters);
	}

	/**
	 * @return the statistics
	 */
	public SimulationStatistics getStatistics() {
		return statistics;
	}


	/**
	 * 
	 */
	public Simulation() {
		eventGen = new EventGenerator();
		logFile = "simulation.sim";//hashCode();
		scheduler = new Scheduler(eventGen);
		eventQueue = new TreeSet<Event>(new Comparator<Event>() {
			@Override
			public int compare(Event e1, Event e2) {
				int compare = (int) (e1.getDate() - e2.getDate());
				if (compare == 0) {
					int i1 = EventGenerator.getAvailableEvents().indexOf(e1.toString());
					int i2 = EventGenerator.getAvailableEvents().indexOf(e2.toString());
					assert i1!=-1 && i2!=-1 : "tostring de event doit renvoyer la chaine presente dans eventgenerator";
					compare = i1 - i2;
				}
				if (compare == 0)
					compare = e1.getName().compareTo(e2.getName());
				if (compare == 0)
					compare = e1.getTask().getId() - e2.getTask().getId();
				if (compare == 0)
					compare = e1.getTask().getPriority() - e2.getTask().getPriority();
				if(compare == 0)
					compare = e1.getProcId() - e2.getProcId();
				if(compare == 0)
					compare = 1;
				return compare;
			}
		});
		for (String criterionName : StatisticsManager.getAvailableStatisticsCriterionNames()) {
			addStatisticsEventListeners(StatisticsManager.getInstance(criterionName));
		}
	}

	private IEventGeneratedListener listener = new IEventGeneratedListener() {
		@Override
		public void actionPerformed(Event event) {
			eventQueue.add(event);
		}
	};

	/**
	 * 
	 */
	private void createInitialActicationEventSet(){
		eventQueue.clear();
		for (ITask t : parameters.getTaskSet()){
			eventGen.generateEvent("activate_task",-1, t, t.getFirstRelease(), null);
		}
	}


	private long date = 0;
	private boolean isLogging;

	/**
	 * 
	 */
	public void simulate(){
		fireStatisticsSimulationStarted();
		Logger.instance().setEnabled(false);
		if(isLogging){
			simulationWriter.open(logFile);
			simulationWriter.writeParameters(parameters);
		}
		eventGen.addEventGeneratedListener(listener);
		createInitialActicationEventSet();
		SimulationEndManager endManager = new SimulationEndManager();

		while (!eventQueue.isEmpty() && !endManager.isFinished && date < parameters.getSimulationEnd()) {

			Event e = eventQueue.pollFirst();
			assert e.getDate() >= date : "back to the future : event " + e + " at time " + date ;
			date = e.getDate();

			if(date <= parameters.getSimulationEnd()){
				fireStatisticsNextEventReady(e);
				performEvent(e, date);
				Logger.instance().log(e,"");
				if(isLogging)
					simulationWriter.writeEvent(e);

				fireEventPerformed(e);
				fireStatisticsEventPerformed(e);
				e.accept(endManager);
			}
			else date = parameters.getSimulationEnd();
			
			// to do not exceed E_max
			long energy = parameters.getEnergyProfile().getCurrentEnergyLevel();
			if(parameters.isEnergyAware() && energy < parameters.getEnergyProfile().getMaxEnergyLevel()){ 
				Event nextEvent = eventQueue.first();
				long d = nextEvent!=null ? nextEvent.getDate() - date : 0;
				for (int i = 0; i < d; i++) {
					energy += parameters.getEnergyProfile().howMuchCanRecharge(date+i, 1);
					if(energy >= parameters.getEnergyProfile().getMaxEnergyLevel()){
						if(energy > parameters.getEnergyProfile().getMaxEnergyLevel() && i>0)
							eventGen.generateEvent("check_energy_state",0, parameters.getTaskSet().getOne(), date+i, null);
						eventGen.generateEvent("check_energy_state",0, parameters.getTaskSet().getOne(), date+i+1, null);
						break;
					}
				}
			}
			statistics.setLastEvent(e);
		}
		eventGen.removeEventGeneratedListener(listener);
		if(isLogging)
			simulationWriter.close();

		fireStatisticsSimulationEnd(date);

		for (IStatisticCriterion criterion : statisticsEventListeners) {
			statistics.addProperty(criterion.getName(), criterion); 
		}
	}



	class SimulationEndManager extends EventVisitorAdapter {
		public boolean isFinished = false;
		@Override public void visit(DeadlineMissedEvent e) {
			if(parameters.isHardRealtime())
				failureEnd(e);
		}
		@Override public void visit(EnergyFailureEvent e) {
			if(parameters.isEnergyAware())
				failureEnd(e);
		}
		private void failureEnd(Event e){
			isFinished = true;
		}
	}

	public void setLogFile(String file) {
		this.logFile = file;
	}

	public void setLoggingEnabled(boolean islogging){
		this.isLogging = islogging;
	}

	private void performEvent(Event event, long date){
		event.actionPerformed(date);
		scheduler.schedule(event, parameters.getTaskSet(), parameters.getSchedulingPolicy(), parameters.getEnergyProfile(), date, parameters.isEnergyAware());
	}

	private List<IEventAddedListener> listeners = new ArrayList<IEventAddedListener>();

	public void addEventAddedListerner(IEventAddedListener l){
		if(!listeners.contains(l))
			listeners.add(l);
	}

	public void removeEventAddListener(IEventAddedListener l) {
		if(listeners.contains(l))
			listeners.remove(l);
	}

	protected void fireEventPerformed(Event e){
		for (IEventAddedListener l : listeners) {
			l.perfomredAction(e);
		}
	}

	private List<IStatisticCriterion> statisticsEventListeners = new ArrayList<IStatisticCriterion>();

	public void addStatisticsEventListeners(IStatisticCriterion l){
		l.setSimulationParameters(parameters);
		if(!statisticsEventListeners.contains(l))
			statisticsEventListeners.add(l);
	}

	public void removeStatisticsEventListeners(IStatisticCriterion l) {
		if(statisticsEventListeners.contains(l))
			statisticsEventListeners.remove(l);
	}

	protected void fireStatisticsSimulationStarted(){
		for (IStatisticCriterion l : statisticsEventListeners)
			l.simulationStarted();
	}
	protected void fireStatisticsSimulationEnd(long date){
		for (IStatisticCriterion l : statisticsEventListeners)
			l.simulationEnd(date);
	}
	protected void fireStatisticsNextEventReady(Event event){
		for (IStatisticCriterion l : statisticsEventListeners)
			l.nextEventReady(event);
	}
	protected void fireStatisticsEventPerformed(Event event){
		for (IStatisticCriterion l : statisticsEventListeners)
			l.eventPerformed(event);
	}






}
