package yartiss.engine.simulation.io.one;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import yartiss.engine.simulation.IEnergyProfile;
import yartiss.engine.simulation.ISchedulingPolicy;
import yartiss.engine.simulation.SimulationParameters;
import yartiss.engine.simulation.energy.profiles.EnergyProfileFactory;
import yartiss.engine.simulation.scheduling.policies.SchedulingPolicyFactory;
import yartiss.engine.simulation.system.ITask;
import yartiss.engine.simulation.system.ITaskSet;
import yartiss.engine.simulation.system.schedulable.GraphTask;
import yartiss.engine.simulation.system.schedulable.SchedulableFactory;
import yartiss.engine.simulation.system.schedulable.SubTask;

public class XmlSimulationParametersReader implements ISimulationParametersReader{

	private SimulationParameters parameters;
	private ISchedulingPolicy policy;
	private int simulationEnd;
	private IEnergyProfile energyProfile;
	private ITaskSet tasks;
	private int nbProc;
	private List<Integer> policyParams;
	private HashMap<String, Object> taskMap, subtaskMap;
	private ITask task;
	private List<SubTask> subtasks;
	private String type;
	private boolean isEnergyAware;
	private boolean isHardRealTime;

	
	@Override
	public SimulationParameters read(File file)throws FileNotFoundException , IOException {
		return read(new FileInputStream(file));
	}

	@Override
	public SimulationParameters read(String file) throws FileNotFoundException , IOException {
		return read(new FileInputStream(file));
	}

	@Override
	public SimulationParameters read(InputStream input) throws IOException {
		try {
			reset();
			SAXParserFactory factory = SAXParserFactory.newInstance();
			factory.setNamespaceAware(true);
			SAXParser parser = factory.newSAXParser();
			parser.parse(input, this.getDefaultHandler());
		} catch (Exception e) {
			throw new IOException(e.getLocalizedMessage());
		}
		return parameters;
	}
	
	private void reset(){
		parameters = null;
		policy = null;
		simulationEnd = 0;
		energyProfile= null;
		tasks = null;
		nbProc = 0;
		policyParams = null;
		taskMap = null;
		subtaskMap = null;
		task = null;
		subtasks = null;
		type = "";
	}

	protected DefaultHandler getDefaultHandler() {
		return new DefaultHandler() {


			@Override
			public void startElement(String uri, String localName,
					String qName, Attributes attributes) throws SAXException {

				if ("policy".equals(localName)) {
					String name = attributes.getValue("name");
					policyParams = new ArrayList<Integer>();
					//nbParams = Integer.parseInt(attributes.getValue("nbParams"));
					policy = SchedulingPolicyFactory.getInstance(name);
				} else if ("parameter".equals(localName)) {
					policyParams.add(Integer.parseInt(attributes
							.getValue("value")));
				} else if ("simulation".equals(localName)) {
					simulationEnd = Integer.parseInt(attributes.getValue("endValue"));
					String value = attributes.getValue("isEnergyAware");
					isEnergyAware = value!=null ? Boolean.parseBoolean(value) : true;
					value = attributes.getValue("isHardRealTime");
					isHardRealTime =value!=null ? Boolean.parseBoolean(value) : true;
					value = attributes.getValue("nbProc");
					nbProc = value != null ? Integer.parseInt(value) : 1;
				} else if ("energyProfile".equals(localName)) {
					int[] energyParams;
					if (attributes.getValue("E0") == null)
						energyParams = new int[3];
					else {
						energyParams = new int[4];
						energyParams[3] = Integer.parseInt(attributes
								.getValue("E0"));
					}
					energyParams[0] = Integer.parseInt(attributes
							.getValue("pr"));
					energyParams[1] = Integer.parseInt(attributes
							.getValue("Emax"));
					energyParams[2] = Integer.parseInt(attributes
							.getValue("Emin"));
					energyProfile = EnergyProfileFactory.getInstance(
							"default", energyParams);
				} else if ("tasks".equals(localName)) {
					if (policy != null){
						tasks = policy.createTaskSet();
						type = attributes.getValue("type");
					}
					else
						throw new IllegalArgumentException(
								"Policy is null and can't create a taskset!");
				} else if ("task".equals(localName)) {
					subtasks = new ArrayList<SubTask>();
					type = attributes.getValue("type");

					taskMap = new HashMap<String, Object>();
					taskMap.put("type", type);
					String param =type.toLowerCase();
					if("simple".equals(param)){

						taskMap.put("firstrelease",
								attributes.getValue("firstRelease"));
						taskMap.put("wcet", attributes.getValue("wcet"));
						taskMap.put("wcee", attributes.getValue("wcee"));
						taskMap.put("period", attributes.getValue("period"));
						taskMap.put("deadline", attributes.getValue("deadline"));
						taskMap.put("priority", attributes.getValue("priority"));
						task = SchedulableFactory.newInstance(taskMap);
					}
					else if( "graph".equals(param)){

						subtasks = new ArrayList<SubTask>(
								Integer.parseInt(attributes
										.getValue("nbSubtasks")));
						taskMap.put("firstrelease",
								attributes.getValue("firstRelease"));
						taskMap.put("period", attributes.getValue("period"));
						taskMap.put("deadline", attributes.getValue("deadline"));
						taskMap.put("priority", attributes.getValue("priority"));

					}
					else if( "subtask".equals(param));
					else 
						throw new IllegalArgumentException(
								"Illegal task type...");

				} else if ("subtask".equals(localName)) {
					subtaskMap = new HashMap<String, Object>();
					subtaskMap.put("type", attributes.getValue("type"));
					subtaskMap.put("firstrelease",
							attributes.getValue("firstRelease"));
					subtaskMap.put("wcee", "0");
					subtaskMap.put("wcet", attributes.getValue("wcet"));
					subtaskMap.put("period", attributes.getValue("period"));
					subtaskMap.put("deadline", attributes.getValue("deadline"));
					subtaskMap.put("priority", attributes.getValue("priority"));
					subtaskMap.put("nbproc", attributes.getValue("nbProc"));

					SubTask subtask = (SubTask) SchedulableFactory
							.newInstance(subtaskMap);
					subtask.setChildrenIndices(attributes.getValue("children")
							.trim().split(" "));
					subtask.setId(Integer.parseInt(attributes.getValue("index")));
					subtasks.add(subtask);

				}
				/*else if ("task".equals(localName)) {
					if (tasks == null)
						throw new IllegalArgumentException("taskset is null!");
					if (!tasks.addTask(SchedulableFactory.newInstance(
							attributes.getValue("type"),
							attributes.getValue("start"),
							attributes.getValue("wcet"),
							attributes.getValue("wcee"),
							attributes.getValue("period"),
							attributes.getValue("deadline"),
							attributes.getValue("priority")))) {
						System.err
								.println("Warning: 2 tasks with the same priority!");
					}
				}*/
				super.startElement(uri, localName, qName, attributes);
			}

			@Override
			public void endElement(String uri, String localName, String qName)
					throws SAXException {

				if ("dataset".equals(localName)) {
					parameters = new SimulationParameters(policy, simulationEnd, energyProfile, tasks, nbProc, isEnergyAware, isHardRealTime);
				}else if("policy".equals(localName)){
					if(policyParams.size() != 0 || policy != null){
						policy.setParameters(policyParams);
					}
				}else if ("task".equals(localName)) {
					if (!subtasks.isEmpty()) {
						taskMap.put("tasks", subtasks);

						task = SchedulableFactory.newInstance(taskMap);
						for (SubTask sub : ((GraphTask) task).getSubtasks()) {
							if (sub.getChildrenIndices() == null)
								continue;
							if (sub.getChildrenIndices().contains(-1))
								continue;
							for (SubTask sub2 : ((GraphTask) task)
									.getSubtasks()) {
								if ((sub.getChildrenIndices().contains(sub2
										.getId())) && !sub.equals(sub2)) {
									sub.addChildTask(sub2);
								}
							}
						}
					}
					if (task != null) {
						tasks.addTask(task);
					}
				}
				super.endElement(uri, localName, qName);
			}

			@Override
			public void characters(char[] ch, int start, int length)
					throws SAXException {
				super.characters(ch, start, length);
			}

		};
	}

}
