package yartiss.engine.simulation.scheduling.policies.fp;

import java.util.SortedSet;

import yartiss.engine.simulation.Event;
import yartiss.engine.simulation.IEnergyProfile;
import yartiss.engine.simulation.ISchedulingPolicy;
import yartiss.engine.simulation.events.EventGenerator;
import yartiss.engine.simulation.scheduling.policies.AbstractMonoProcessorSchedulingPolicy;
import yartiss.engine.simulation.scheduling.policies.tools.FPSlackTimeManager;
import yartiss.engine.simulation.system.ITask;
import yartiss.engine.simulation.system.ITaskSet;

/**
 * Preemptive Fixed Priority. When there is not enough energy, the system is
 * paused for a duration equal to the slack time. However if there is no
 * slack time, the system is paused 1 time unit. (And will probably fail).
 */
public class EHPFP3 extends AbstractMonoProcessorSchedulingPolicy  {

	
	private final FPSlackTimeManager slackTimeManager;
	
	public EHPFP3(String tasksetType, String policyName, FPSlackTimeManager slackTimeManager) {
		super(tasksetType, policyName);
		this.slackTimeManager = slackTimeManager;
	}

	public EHPFP3(FPSlackTimeManager slackTimeManager){
		this("priority","EH_PFP_3", slackTimeManager);
	}
	

	long sleepUntil = -1;
	@Override
	public ITask chooseNextTask(ITaskSet taskSet, IEnergyProfile energyProfile,	long date, EventGenerator evGen) {
		getMetrics().clear();
		SortedSet<ITask> activeTasks = taskSet.getActiveTasks(date);

		if (date == 0)
			sleepUntil = -1;

		if (!activeTasks.isEmpty() && date >= sleepUntil) {
			ITask first = activeTasks.first();
			if (date == sleepUntil && energyProfile.getCurrentEnergyLevel() < energyProfile.getMaxEnergyLevel()) {
				long st = slackTimeManager.getSlackTime(date, taskSet);
				getMetrics().put(Event.SlackTime, st);
				if (st > 0) {
					long x = Math.max(st, 1);
					long de = energyProfile.getMaxEnergyLevel() - energyProfile.getCurrentEnergyLevel();
					long dt = de / energyProfile.getPowerRate();

					x = Math.min(x, dt);

					sleepUntil = date + x;
					if (date != sleepUntil) {
						evGen.generateEvent("check_energy_state", activeTasks.first(), sleepUntil, null);
						return null;
					}
				}
			}

			long hlcet = energyProfile.howLongCanExecute(first);

			if (hlcet <= 0) {
				long st = slackTimeManager.getSlackTime(date, taskSet);
				getMetrics().put(Event.SlackTime, st);
				st = Math.max(st, 1);
				sleepUntil = date + st;
				evGen.generateEvent("check_energy_state", activeTasks.first(), sleepUntil, null);
				return null;
			}

			if (hlcet < activeTasks.first().getRemainingCost())
				evGen.generateEvent("check_energy_state", activeTasks.first(), date + hlcet, null);
			return activeTasks.first();
		}
		return null;
	}

	@Override
	public ISchedulingPolicy newInstance() {
		return new EHPFP3(tasksetType, policyName, new FPSlackTimeManager());
	}

	
}
