package yartiss.engine.simulation.scheduling.policies.fp;

import java.util.SortedSet;

import yartiss.engine.simulation.Event;
import yartiss.engine.simulation.IEnergyProfile;
import yartiss.engine.simulation.ISchedulingPolicy;
import yartiss.engine.simulation.events.EventGenerator;
import yartiss.engine.simulation.scheduling.policies.AbstractMonoProcessorSchedulingPolicy;
import yartiss.engine.simulation.scheduling.policies.tools.FPSlackTimeManager;
import yartiss.engine.simulation.system.ITask;
import yartiss.engine.simulation.system.ITaskSet;

/**
 * Preemptive Fixed Priority. When there is not enough energy, the system is
 * paused for a duration equal to the slack time. However if there is no slack
 * time, the system is paused 1 time unit. (And will probably fail). Morevover,
 * if energy state reaches Emax, the schedule is resumed
 */
public class PFPSlackTime extends AbstractMonoProcessorSchedulingPolicy{

	private final FPSlackTimeManager slackTimeManager;

	public PFPSlackTime(String tasksetType, String policyName, FPSlackTimeManager slackTimeManager) {
		super(tasksetType, policyName);
		this.slackTimeManager = slackTimeManager;
	}
	
	public PFPSlackTime(FPSlackTimeManager slackTimeManager) {
		this("priority", "PFP_ST", slackTimeManager);
	}

	long sleepUntil = -1;
	long sleepUntilDecision = 0;

	@Override
	public ITask chooseNextTask(ITaskSet taskSet, IEnergyProfile energyProfile, long date, EventGenerator evGen) {
		metrics.clear();

		if (date == sleepUntilDecision)
			sleepUntil = -1;

		SortedSet<ITask> activeTasks = taskSet.getActiveTasks(date);
		
		if (!activeTasks.isEmpty()) {
			ITask first = activeTasks.first();

			if (date < sleepUntil) {// on doit laisser le systeme se recharger
			
				// Au cas ou le st a change (devrait pas arriver)
				long st = Math.max(0, slackTimeManager.getSlackTime(date, taskSet));

				metrics.put(Event.SlackTime, st);
				if(st<=0){
					sleepUntilDecision=date;
					evGen.generateEvent("check_energy_state", activeTasks.first(), date, null);
				}
				// fin du au cas ou
				//System.out.println(date+" : 1sleep="+sleepUntil);
				return null;
			}
			if (date == sleepUntil) {// on doit recalculer le slack time,
										// modifier sleepUntil et recommencer

				long de = energyProfile.getMaxEnergyLevel() - energyProfile.getCurrentEnergyLevel();
				if (de < energyProfile.getPowerRate()) {// on reprend
														// l'execution
					sleepUntilDecision = date;
					sleepUntil = -1; // inutile mais pour clarifier le code
					evGen.generateEvent("check_energy_state", activeTasks.first(), date, null);
					//System.out.println(date+" : 2sleep="+sleepUntil);
					return null;
				}
				long st = Math.max(0, slackTimeManager.getSlackTime(date, taskSet));
				
				metrics.put(Event.SlackTime, st);
				// if(st>0){
				long dt = Math.min(de / energyProfile.getPowerRate(), st);
				evGen.generateEvent("check_energy_state", activeTasks.first(), date + dt, null);
				sleepUntilDecision = date;
				sleepUntil = date + dt;
				//System.out.println(date+" : 3sleep="+sleepUntil);
				return null;
				// } La suite est commente car revient au meme
				// sleepUntilDecision=date;
				// sleepUntil=-1; // inutile mais pour clarifier le code
				// evGen.generateEvent("check_energy_state",
				// activeTasks.first(),date, null);
				// return null;
			}
			if (date > sleepUntil) {// on reprend l'execution si c'est possible

				long hlcet = energyProfile.howLongCanExecute(first);
				if (hlcet > 0) {
					if (hlcet < activeTasks.first().getRemainingCost())
						evGen.generateEvent("check_energy_state", activeTasks.first(), date + hlcet, null);
					return activeTasks.first();
				}
				long de = energyProfile.getMaxEnergyLevel() - energyProfile.getCurrentEnergyLevel();
				
				long st = Math.max(0, slackTimeManager.getSlackTime(date, taskSet));
				
				metrics.put(Event.SlackTime, st);

				if(st<=0){// pour que la tache depasse son echeance 
					evGen.generateEvent("check_energy_state", activeTasks.first(), date + 1, null);
					//System.out.println(date+" : 4sleep="+sleepUntil);
					return null;
				}
				
				long dt = Math.min(Math.max(1,de / energyProfile.getPowerRate()), st);
				evGen.generateEvent("check_energy_state", activeTasks.first(), date + dt, null);
				sleepUntilDecision = date;
				sleepUntil = date + dt;
				//System.out.println(date+" : 5sleep="+sleepUntil);
				return null;

			}

		}
		
		long de = energyProfile.getMaxEnergyLevel() - energyProfile.getCurrentEnergyLevel();
		long st = Math.max(0, slackTimeManager.getSlackTime(date, taskSet));
		
		metrics.put(Event.SlackTime, st);
		// if(st>0){
		long dt = Math.min(de / energyProfile.getPowerRate(), st);
		if(dt!=0)
			evGen.generateEvent("check_energy_state", taskSet.getOne(), date + dt, null);
		sleepUntilDecision = date;
		sleepUntil = date + dt;
		//System.out.println(date+" : 6sleep="+sleepUntil+" st="+st);
		return null;
	}

	@Override
	public ISchedulingPolicy newInstance() {
		return new PFPSlackTime(tasksetType, policyName, new FPSlackTimeManager());
	}
}
