package yartiss.engine.simulation.system;

import java.util.Comparator;
import java.util.SortedSet;
import java.util.TreeSet;

import yartiss.engine.simulation.system.schedulable.SchedulableFactory;
import yartiss.engine.simulation.system.schedulable.SimpleTask;
import yartiss.engine.simulation.system.taskset.TaskSetFactory;

public class SystemMetrics {

	private static final Comparator<ITask> comparator = new Comparator<ITask>() {
		@Override
		public int compare(ITask t1, ITask t2) {
			return (int) (t1.getPriority() - t2.getPriority());
		}
	};

	/**
	 * 
	 * @param taskset
	 * @return
	 */
	public static double utilisation(Iterable<ITask> taskset){
		double u = 0;
		double C,T;
		for (ITask task : taskset) {
			C = task.getWcet();
			T = task.getPeriod();
			u += C/T; 
		}
		return u;
	}

	public static double utilisation(ITask[] taskset){
		double u = 0;
		double C,T;
		for (ITask task : taskset) {
			C = task.getWcet();
			T = task.getPeriod();
			u += C/T; 
		}
		return u;
	}
	public static double energyUtilization(ITask[] taskset, long pr){
		double u = 0;
		double Ei,Ti;
		for (ITask task : taskset) {
			Ei = task.getWcee();
			Ti = task.getPeriod();
			u += Ei/Ti; 
		}
		return u/pr;
	}
	public static double energyUtilization(Iterable<ITask> taskset, long pr){
		double u = 0;
		double Ei,Ti;
		for (ITask task : taskset) {
			Ei = task.getWcee();
			Ti = task.getPeriod();
			u += Ei/Ti; 
		}
		return u/pr;
	}

	public static boolean  isEnergyFaisible(ITask [] taskSet,long pr){
		ITaskSet ts = TaskSetFactory.getNewInstance("priority");
		for (ITask task : taskSet) {
			ts.addTask(task);
		}
		double Ue = energyUtilization(ts, pr);
		return Ue <= 1;
	}


	/**
	 * 
	 * @param taskset
	 * @return
	 */

	public static boolean isFPTimeFaisible(ITask [] taskSet){
		ITaskSet ts = TaskSetFactory.getNewInstance("priority");
		for (ITask task : taskSet) {
			ts.addTask(task);
		}
		return isFPTimeFaisible(ts);
	}

	public static boolean isFPTimeFaisible(ITaskSet ts){
		ITaskSet taskSet = TaskSetFactory.getNewInstance("priority");
		for (ITask task : ts) {
			taskSet.addTask(task);
		}
		double u = utilisation(taskSet);
		if(u > 1)return false;
		switch (detectDeadLineModel(taskSet)) {
		case CONSTRAINED_DEADLINE:
		case IMPLICITE_DEADLINE  :	
			SortedSet<ITask> tasks = new TreeSet<ITask>(comparator);
			for (ITask task : taskSet) 
				tasks.add(task);

			long Ci, Tj,Di,Wt,t,nbInterfer;
			for (ITask task_i : tasks) {
				Ci  = task_i.getWcet();
				Di  = task_i.getDeadline();
				Wt = 0;
				do{
					t = Wt;
					Wt =  Ci;
					for (ITask task_j : tasks.headSet(task_i)) {
						Tj  = task_j.getPeriod();
						nbInterfer  = t / Tj;
						nbInterfer += (t % Tj ==0 ? 0 : 1);
						Wt += nbInterfer *task_j.getWcet();
					}
				}while(Wt != t && Wt <= Di);
				if(Wt > Di)return false;
			}
			return true;
		}
		// TODO arbitrary deadline
		return false;
	}	


	/**
	 * 
	 * @param tasks
	 * @return
	 */
	public static DeadlineModel detectDeadLineModel(ITaskSet tasks){
		boolean implicite = true;
		for (ITask task : tasks) {
			if(task.getDeadline() > task.getPeriod()) return  DeadlineModel.ARBITRARY_DEADLINE;
			if(task.getDeadline() < task.getPeriod())implicite = false;
		}
		return implicite ? DeadlineModel.IMPLICITE_DEADLINE : DeadlineModel.CONSTRAINED_DEADLINE;
	}

	/**
	 * 
	 * @param taskset
	 * @return
	 */

	public static long FPBusyPeriod(Iterable<ITask> taskset){
		ITaskSet ts = TaskSetFactory.getNewInstance("priority");
		for (ITask task : taskset) {
			ts.addTask(task);
		}
		return FPBusyPeriod(ts);
	}

	public static long FPBusyPeriod(ITaskSet taskset){
		ITask lowerPriorityTask = null;
		for (ITask task : taskset) {
			if(lowerPriorityTask==null)lowerPriorityTask = task;
			if(task.getPriority() > lowerPriorityTask.getPriority())
				lowerPriorityTask = task;
		}
		if(lowerPriorityTask != null) return FPBusyPeriodForLevelK(taskset, lowerPriorityTask);
		return -1;
	}

	public static long FPBusyPeriodForLevelK(ITaskSet taskset, ITask task_k){
		long t, w = task_k.getWcet();
		long hp = hyperPeriod(taskset);
		do {
			t = w;
			w = 0;
			for (ITask task_i : taskset.getHeadSet(task_k, true)) {
				w += (t / task_i.getPeriod()) * task_i.getWcet();
				w += (t % task_i.getPeriod() == 0 ? 0 : task_i.getWcet());  
			}
		}while(t!=w && t<=hp);
		return t;
	}

	public static long FPEnergyBusyPeriod(ITaskSet taskset, long pr){
		ITask lowerPriorityTask = null;
		for (ITask task : taskset) {
			if(lowerPriorityTask==null)lowerPriorityTask = task;
			if(task.getPriority() > lowerPriorityTask.getPriority())
				lowerPriorityTask = task;
		}
		if(lowerPriorityTask != null) return FPEnergyBusyPeriodForLevelK(taskset, lowerPriorityTask, pr);
		return -1;

	}

	public static long FPEnergyBusyPeriodForLevelK(ITaskSet taskset, ITask task_k, long pr){
		long t, w = task_k.getWcet(), we = task_k.getWcee();
		do {
			t = w;
			w = 0;
			we=0;
			for (ITask task_i : taskset.getHeadSet(task_k, true)) {
				long d = t / task_i.getPeriod();
				d += (t % task_i.getPeriod() == 0 ? 0 : 1); 
				w += d * task_i.getWcet();
				we += d * task_i.getWcee();
			}
		}while(t!=w);
		long needed = we/pr;
		needed += (we % pr == 0 ? 0 : 1); 
		return Math.max(needed, w);
	}

	public static boolean isTimeAndEnergyFaisible(){
		return false;
	}


	private static long  lcm (long nb1, long nb2) {
		long prod, rest, lcm;

		prod = nb1*nb2;
		rest   = nb1%nb2;
		while(rest != 0){
			nb1 = nb2;
			nb2 = rest;
			rest = nb1%nb2;
		}
		lcm = prod/nb2;
		return lcm;		
	} 

	private static long lcm(long nb1, long ... nbs){
		long lcm = nb1;
		for (long nbi : nbs) {
			lcm = lcm(lcm, nbi);
		}
		return lcm;
	}

	public static long hyperPeriod(Iterable<ITask> taskset){
		int count = 0;
		for (@SuppressWarnings("unused") ITask task : taskset) count++;
		long [] tasksPeriods = new long [count];
		int i = 0;
		long p1 = 0;
		for (ITask task : taskset) {
			tasksPeriods[i] = task.getPeriod();
			p1 = task.getPeriod();
			i++;
		}
		return lcm(p1,tasksPeriods);
	}

	public static long effectiveDeadline(ITask task, ITaskSet taskset, long di) {
		long edi = di, lastedi;
		long t = di - task.getDeadline();
		//System.out.println("#### tk="+task.getPriority()+"### t="+t+" #### di="+di+" ############");
		do{
			lastedi = edi;
			for (ITask task_j : taskset.getHeadSet(task, false)) {
				long xj = task_j.getNextActivation(t);
				long fj = Math.max(0, (edi-xj)/task_j.getPeriod());
				if(xj<edi && xj + fj * task_j.getPeriod() + task_j.getWcet() >= edi ){
					edi = xj + fj*task_j.getPeriod();
					//System.out.println("t"+task_j.getPriority()+" edi="+edi+" fj="+fj+" xj="+xj+" ci="+task_j.getWcet());
				}
			}
			//System.out.println("^^^^^^^^^^^^^^^");
		}while(edi != lastedi);
		return edi;

	}










	public static void main(String[] args) {
		ITaskSet set = TaskSetFactory.getNewInstance("priority");
		//set.addTask(new SimpleTask(firstRelease, wcet, wcee, period, deadline, priority, p));
		set.addTask(new SimpleTask(0, 2, 32, 84, 73, 1,null));
		set.addTask(new SimpleTask(0, 1, 16, 35, 5, 2,null));
		set.addTask(new SimpleTask(0, 2, 34, 60, 42, 3,null));
		set.addTask(new SimpleTask(0, 3, 48, 20, 3, 4,null));
		
		System.out.println(isFPTimeFaisible(set));
		
	}











}
