package yartiss.engine.simulation.system.taskset.generation;

import java.util.Arrays;
import java.util.Comparator;
import java.util.Random;
import java.util.TreeSet;

import yartiss.engine.simulation.io.many.ITasksetsWriter;
import yartiss.engine.simulation.io.many.TasksetsIOFactory;
import yartiss.engine.simulation.system.ITask;
import yartiss.engine.simulation.system.SystemMetrics;
import yartiss.engine.simulation.system.schedulable.SchedulableFactory;

public class FrameBasedUUnifastDiscardTaskSetGenerator implements ITaskSetGenerator{

	private double Ue;
	private int nbTask;
	private double U;
	private boolean offset;
	private int deadlineModel;
	private long Emax;
	private long pr;
	private int nbProc = 1;
	private Random random = new Random(System.currentTimeMillis());


	public FrameBasedUUnifastDiscardTaskSetGenerator( int nbTask, int nbProc, double utilization, double energyUtilization, long Pr,
			long emax, int deadlineModel, boolean offset) {

		if (utilization > 1)throw new IllegalArgumentException(	"Utilization must be less than or equal to the number of processors.");
		if (energyUtilization > 1) 	throw new IllegalArgumentException("Energy Utilization must be less than or equal to 1.");

		this.Ue = energyUtilization;
		this.U = utilization;
		this.offset = offset;
		this.deadlineModel = deadlineModel;
		this.Emax = emax;
		pr = Pr;
		this.nbTask = nbTask;

	}
	@Override
	public String getName() {
		return "FameBasedGenerator";
	}

	@Override
	public String getDescription() {
		return "FameBased";
	}

	@Override
	public ITask[] generate() {

		ITask[] taskset = new ITask[nbTask];
		do{
			long period = generatePeriod();
			
			long deadline = period; // implicit deadline

			if (deadlineModel < 0) { // constrained deadline
				deadline = nextInt(random, (int) (U*period), (int) period);
			} else if (deadlineModel > 0) { // arbitrary deadline
				deadline = nextInt( random, 1, MAX_PERIOD);
			}
			
			
			double[] util = generateUtilizations();
			double[] energyUtil = generatEnergyUtilization(util);
			for (int i = 0; i < util.length; i++) {

				taskset[i] = generate(period, deadline,  util[i],energyUtil[i], pr,Emax, offset,deadlineModel, i==util.length-1);


				double u = ((double)taskset[i].getWcet())/taskset[i].getPeriod();
				double diff = u - util[i];
				if(i+1 < util.length)
					util[i+1] -= diff;

				double ue = ((double)taskset[i].getWcee()) / (taskset[i].getPeriod() * pr);
				diff = ue - energyUtil[i];
				if(i+1 < energyUtil.length)
					energyUtil[i+1] -= diff;
			}
		}while(!SystemMetrics.isFPTimeFaisible(taskset) || !SystemMetrics.isEnergyFaisible(taskset, pr));


		int priority=1;
		for (ITask t : taskset) {
			t.setPriority(priority++);

			long d = t.getWcee() / t.getWcet() - pr;
			if( d > Emax)
				throw new IllegalStateException("E/C-pr <=0 : "+t.getWcee()+"/"+t.getWcet()+"-"+pr+"="+d+" <="+Emax);
//			
//			if(ue-u > 0.01){
//				double ui = ((double)t.getWcet()) / t.getPeriod();
//				double uei= ((double)t.getWcee()) / (t.getPeriod()*pr);
//				if( uei <= ui)
//					throw new IllegalStateException("uei <= ui : "+uei+"<="+ui);
//				
//				if( d <= 0)
//					throw new IllegalStateException("E/C-pr <=0 : "+t.getWcee()+"/"+t.getWcet()+"-"+pr+"="+d);
//				
//			}
		}
		
		return taskset;
	}

	private long generatePeriod(){

		int maxHyperperiod = 200;
		int from = (int) Math.max(nbTask/(Ue*pr), nbTask/U);
		long period = nextInt(random, from, maxHyperperiod);
		return period;
	}
	
	

	/**
	 * 
	 * @return
	 */
	private double[] generateUtilizations() {
		Random random = new Random(System.currentTimeMillis());
		double[] util = new double[nbTask];
		double nextSumU;
		boolean discard;

		do {
			double sumU = this.U;
			discard = false;
			for (int i = 0; i < nbTask - 1; i++) {
				nextSumU = sumU	* Math.pow(random.nextDouble(), (double) 1	/ (nbTask - (i + 1)));
				util[i] = sumU - nextSumU;
				sumU = nextSumU;
				if (util[i] > 1) {
					discard = true;
				}
			}
			util[nbTask - 1] = sumU;
			if (util[nbTask - 1] > 1) {
				discard = true;
			}
		} while (discard || !utilizationIsValid(util));
		return util;
	}

	/**
	 * 
	 * @param util
	 * @return
	 */
	private boolean utilizationIsValid(double[] util) {
		double sum = 0;

		for (double u : util) {
			sum += u;
		}
		return sum <= nbProc ? true : false;
	}

	/**
	 * 
	 * @return
	 */
	private double[] generatEnergyUtilization(double[] u) {
		Random random = new Random(System.currentTimeMillis());
		double[] ue = new double[nbTask];
		double nextSumU;
		boolean discard;

		do {

			double sumU = this.Ue;
			discard = false;
			for (int i = 0; i < nbTask - 1; i++) {
				nextSumU = sumU	* Math.pow(random.nextDouble(), (double) 1	/ (nbTask - (i + 1)));
				ue[i] = sumU - nextSumU;
				sumU = nextSumU;
				if (ue[i] > 1	|| (ue[i] / u[i]) > (((double) Emax) / pr + 1)) {
					discard = true;
				}
			}
			ue[nbTask - 1] = sumU;
			if (ue[nbTask - 1] > 1) {
				discard = true;
			}
		} while (discard || !utilizationIsValid(ue));
		return ue;
	}




	final static int MAX_PERIOD = 1000;
	private static int CURRENT_PRIORITY = 1;

	public ITask generate(long period,long deadline, double utilization, double energyUtilization,	long Pr,long Emax, boolean offset, int deadlineModel, boolean isLast) {
		Random random = new Random();
		long start = 0, wcet = 0, wcee = 0, x;

		wcet = Math.max(1, (long) (period * utilization));
		utilization = ((double) wcet) / period;

		x = Math.round( (energyUtilization / utilization - 1) * Pr); //Ei=(Pr+x)*Ci
		if(x > Emax)
			x=Emax;

		if(Ue > U || Pr+x < 1){
			x = Math.max(1, x);
		}

		wcee = (Pr + x) * wcet;

		int priority = CURRENT_PRIORITY++;
		if (CURRENT_PRIORITY > 100)
			CURRENT_PRIORITY = 1;

		if (offset) {
			start = nextInt(random, 0, (int) period);
		}

		return SchedulableFactory.newInstance("simple", start, wcet, wcee, period, deadline, priority);
	}

	protected int nextInt(Random random, int from, int to) {
		return from + random.nextInt(to - from + 1);
	}

	//test main

	public static void main(String[] args) throws Exception {
		
		int n=20 ,nb=5;
		double start = 0.2;

		int pr = 15;
		int nbTasks = 4;
		int counter = 0;
		double unit = 1./n;
		long s = Math.round(start/unit);
		int maxIndex = n;
		long count =  (maxIndex+1-s)*(maxIndex+1-s)*nb;
		int emax = 200;
		System.out.println(s);
		//System.exit(0);
		
		String destFile = System.getProperty("user.home");// + "/Desktop";
		destFile = destFile.replace("\\", "/")+"/generation_frame_based_"+emax+"_"+pr+"_"+nbTasks+"_"+unit+".txt";
		System.out.println(destFile);
		TreeSet<ITask[]> [][] stats = new TreeSet[n+1][n+1];

		for (int i = 0; i < stats.length; i++)
			for (int j = 0; j < stats.length; j++)
				stats[i][j]=new TreeSet<ITask[]>(new Comparator<ITask[]>() {
					@Override
					public int compare(ITask[] set1, ITask[] set2) {
						if(set1.length != set2.length)return 1;
						for (int i = 0; i < set1.length; i++) {
							if ( 	set1[i].getFirstRelease() != set2[i].getFirstRelease()
									||	set1[i].getWcet() != set2[i].getWcet()
									||	set1[i].getWcee() != set2[i].getWcee()
									||	set1[i].getPeriod() != set2[i].getPeriod()
									||	set1[i].getDeadline() != set2[i].getDeadline()
									//||	set1[i].getPriority() != set2[i].getPriority()
									)
								return 1;
						}
						return 0;
					}
				});

				int oldCounter = 0;
				int maxIterations = 10000;
				int iterationCounter = 0;
				do{
					iterationCounter++;
					for (long i = s; i <= maxIndex; i++) {
						for (long j = s; j <= maxIndex; j++) {
							for (int k = 0; k < nb; k++) {
								double u  =  Math.round((unit*i) * 1000) / 1000.0d;
								double ue =  Math.round((unit*j) * 1000) / 1000.0d;
								//System.out.println("---- u="+u+"	ue="+ue);

								int line = (int) Math.round(u / unit);
								int row = (int) Math.round(ue / unit);
								
								if(stats[line][row].size() < nb){
									FrameBasedUUnifastDiscardTaskSetGenerator gen = new FrameBasedUUnifastDiscardTaskSetGenerator(nbTasks,1, u, ue, pr, emax, 0, false);
									ITask [] taskset = gen.generate();

									u  = Math.round(SystemMetrics.utilisation(Arrays.asList(taskset)) * 1000) / 1000.0d;
									ue = Math.round(SystemMetrics.energyUtilization(Arrays.asList(taskset), pr) * 1000) / 1000.0d;

									line = (int) Math.round(u / unit);
									row = (int) Math.round(ue / unit);

									if(line >= s && row >= s &&  line <= maxIndex    &&    row <= maxIndex    &&   u<=1 && ue<=1   &&   stats[line][row].size() < nb){

										int oldSize = stats[line][row].size();
										stats[line][row].add(taskset);

										if(oldSize < stats[line][row].size()){
											oldCounter = counter;
											counter++;
										}
										else {
											//System.err.println("doublon supprimé");
										}
									}
								}
							}
						}
					}
					
					//if(counter > oldCounter)
					//	flush(stats, destFile, pr);
					
					if(iterationCounter % maxIterations==0)
						nbTasks = nbTasks==2 ? nbTasks : nbTasks-1;
					
					//System.out.println("("+line+" , "+row+")"+"    u="+u+"	ue="+ue+"  hp="+SystemMetrics.hyperPeriod(Arrays.asList(taskset)));
					System.out.println("---------------------------------------"+counter+"/"+count);

					showTable(stats, unit);
					
					
				}while(counter<count);

				flush(stats, destFile, pr);
	}
	
	private static void flush(TreeSet<ITask[]> [][] stats, String destFile, long pr) throws Exception{
		
		final ITasksetsWriter writer = TasksetsIOFactory.newTasksetsWriterInstance("txt",destFile, pr);
		for (int i = 0; i < stats.length; i++){
			for (int j = 0; j < stats.length; j++){
				for (ITask[] taskset : stats[i][j]) {
					writer.write(taskset);
				}
				stats[i][j].clear();
				stats[i][j] = null;
			}
		}
		writer.close();
		System.out.println("writting on :"+destFile);
	}
	
	
	private static void showTable(TreeSet<ITask[]> [][] stats, double unit){
		System.out.print("ue/e\t");
		for (int p = 0; p < stats.length; p++) {
			System.out.print(Math.round(p*unit * 1000)/1000.0d  +"\t");
		}
		System.out.println();
		for (int p = 0; p < stats.length; p++) {
			System.out.print(Math.round(p*unit * 1000)/1000.0d +"\t");
			for (int m = 0; m < stats[0].length; m++) {
				System.out.print(stats[p][m].size()+"\t");
			}
			System.out.println();
		}
	}
	
	
	

}
