package yartiss.engine.simulation.system.taskset.generation;

import java.io.IOException;
import java.util.Arrays;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Callable;
import java.util.concurrent.LinkedBlockingQueue;

import yartiss.engine.simulation.IEnergyProfile;
import yartiss.engine.simulation.ISchedulingPolicy;
import yartiss.engine.simulation.Simulation;
import yartiss.engine.simulation.SimulationParameters;
import yartiss.engine.simulation.energy.profiles.EnergyProfileFactory;
import yartiss.engine.simulation.io.many.ITasksetsWriter;
import yartiss.engine.simulation.io.many.TasksetsIOFactory;
import yartiss.engine.simulation.scheduling.policies.SchedulingPolicyFactory;
import yartiss.engine.simulation.system.ITask;
import yartiss.engine.simulation.system.ITaskSet;
import yartiss.engine.simulation.system.SystemMetrics;
import yartiss.engine.simulation.system.taskset.StaticDeadlineMonotonicTaskSet;
import yartiss.engine.simulation.system.taskset.TaskSetFactory;
import yartiss.engine.simulation.system.taskset.generation.task.DefaultTaskGenerator;
import yartiss.engine.simulation.system.taskset.generation.task.GraphTaskGenerator;
import yartiss.framework.concurrency.Consumer;
import yartiss.framework.concurrency.IElementConsumedListener;
import yartiss.framework.concurrency.MultiThreadProducer;
import yartiss.framework.concurrency.QueueEntry;


public class SystemsGenerator {

	private static int counter = 0;
	private static boolean finished;

	/**
	 * Generates tasksets with specific types of task
	 * 
	 * @param count
	 *            number of tasksets
	 * @param nbProc
	 *            number of processors
	 * @param nbTasks
	 *            number of tasks in each taskset
	 * @param pr
	 *            power rate
	 * @param emax
	 *            maximum energy
	 * @param file
	 *            output file
	 * @param taskType
	 *            type of task (simple, graph, ...)
	 * 
	 * @throws IOException
	 * @throws Exception
	 */
	public static void generateTaskSets(final int count, int nbProc,
			int nbTasks, final int pr, final long emax, String file,
			final String taskType) throws Exception {
		finished = false;
		boolean energy;
		double padUnit = 0.05, pad = 0;
		long n, nb;
		do {
			pad += padUnit;
			n = Math.round(1 / pad);
			nb = count / (n * n);
		} while (nb <= 0);
		if (pad >= 1) {
			pad = 0.9;
			n = 1;
			nb = count;
		}
		// System.out.println(n+"*"+n+"*"+nb+"  pad="+pad);
		counter = 1;
		// final TasksetWriter writer = new TasksetWriter(file);
		// final ITaskElementVisitor writer = new TextWriter(file, count,
		// nbTasks);
		final ITasksetsWriter writer = TasksetsIOFactory.newTasksetsWriterInstance("xml", file, pr);
		BlockingQueue<QueueEntry<ITask[]>> queue = new LinkedBlockingQueue<QueueEntry<ITask[]>>(100000);

		final Consumer<ITask[]> consumer = new Consumer<ITask[]>(queue,
				new IElementConsumedListener<ITask[]>() {
					@Override
					public void consume(ITask[] tasks) {
						try {
							if (tasks.length < 1 || counter > count) {
								System.out.println("hehe");
								return;
							}

							double u = Math.round(SystemMetrics
									.utilisation(Arrays.asList(tasks)) * 1000) / 1000.0d;
							double ue = Math.round(SystemMetrics
									.energyUtilization(Arrays.asList(tasks), pr) * 1000) / 1000.0d;

							System.out.println(counter + "\t: U=" + u + "\tUe="
									+ ue);
							// writer.writeTaskSet(tasks, pr);
							// writer.visitTaskset(Arrays.asList(tasks), pr);
							writer.write(tasks);
							// System.out.print("\033[H"+(counter*100)/count+"          %");
							if (counter >= count) {
								finished = true;
								System.out.println("writer closed!");
								writer.close();
							}
							counter++;

						} catch (IOException e) {
							e.printStackTrace();
						}
					}
				});
		consumer.start();
		do {
			final MultiThreadProducer<ITask[]> producer = new MultiThreadProducer<ITask[]>(queue);
			for (int i = 1; i <= n; i++) {
				for (int j = 1; j <= n; j++) {
					if (i * pad < 1 && j * pad < 1) {

						ITaskGenerator generator;
						if (taskType.toLowerCase().equals("simple")) {
							generator = new DefaultTaskGenerator();
							energy = true;
						} else if (taskType.toLowerCase().equals("graph")) {
							System.out.println("prodiucer:; graph");
							generator = new GraphTaskGenerator();
							energy = false;
						} else
							throw new IllegalArgumentException("Task model ("
									+ taskType + ") doesn't exist...");

						final ITaskSetGenerator taskSetGenerator = new UUniFastDiscardTaskSetGen(
								generator, nbTasks, nbProc, i * pad, j * pad,
								pr, emax, 0, false, energy);
						for (int k = 0; k < nb; k++) {
							producer.produce(new Callable<ITask[]>() {
								@Override
								public ITask[] call() throws Exception {

									if (finished) {
										return null;
									}

									ITask[] tasks = taskSetGenerator.generate();
									if (tasks == null) {
										System.out.println("NULL");
										return new ITask[] {};
									}
									for (ITask iTask : tasks) {
										System.out.println("ITASK: "
												+ iTask.toString());
									}
									return tasks;
								}
							});
						}
					}
				}
			}
			Thread.sleep(1000);
			producer.awaitTermination();
		} while (counter < count);
		queue.put(new QueueEntry<ITask[]>(null));
		System.out.println("fini");
	}

	public static void generateTaskSets(int count, int nbProc, int nbTasks,
			double U, double Ue, int pr, long emax, String file, String taskType)
					throws Exception {
		ITaskSetGenerator taskSetGenerator = null;
		// TasksetWriter writer = new TasksetWriter(file);
		taskSetGenerator = new UUniFastDiscardTaskSetGen(
				new DefaultTaskGenerator(), nbTasks, nbProc, U, Ue, pr, emax,
				0, false, true);
		for (int i = 0; i < count; i++) {
			ITask[] tasks = taskSetGenerator.generate();
			for (ITask task : tasks) {
				System.out.println(task);
			}
			System.out
			.println(i
					+ " ## U="
					+ SystemMetrics.utilisation(Arrays.asList(tasks))
					+ "##Ue="
					+ SystemMetrics.energyUtilization(
							Arrays.asList(tasks), pr));
			ITasksetsWriter writer = TasksetsIOFactory.newTasksetsWriterInstance("txt",file,pr);
			// writer.visitTaskset(Arrays.asList(tasks), pr);
			writer.write(tasks);
			// writer.writeTaskSet(tasks, pr);
		}
		// writer.close();
	}

	public static void oldGenerate(Class<?> clazz) throws Exception {
		String setsDest = "systems/old_";
		String setsSrc = "/resources/tasks_sets";
		String energySrc = "/resources/energy_";

		long[] pr = new long[] { 10, 15, 20, 25, 30, 35 };
		long[] emaxs = new long[] { 200, 400 };
		for (int i = 0; i < emaxs.length; i++) {
			for (int j = 0; j < pr.length; j++) {
				ITasksetsWriter writer = TasksetsIOFactory.newTasksetsWriterInstance("txt",setsDest + emaxs[i]+ "_" + pr[j] + ".txt",pr[j]);
				FileTaskSetGenerator generator = new FileTaskSetGenerator(
						clazz.getResourceAsStream(setsSrc),
						clazz.getResourceAsStream(energySrc + emaxs[i] + "_"
								+ pr[j] + ".txt"), pr[j]);
				int count = 30;// generator.getSetsCount();
				for (int k = 0; k < count; k++) {
					writer.write(generator.generate());
				}
				writer.close();
			}
		}
	}

	public static void main(String[] args) throws Exception {
		int count = 100; 
		int nbProc = 1;
		int nbTasks = 0;
		double U = 0.8; 
		double Ue = 0.2; 
		int pr = 15; 
		int emax = Integer.MAX_VALUE;
		boolean isEnergyAware = true;
		boolean isHardRealtime = true;
		int simEnd = 100000; 
		TaskSetFactory.registerPolicy(new StaticDeadlineMonotonicTaskSet(), "priority");
		System.out.println("start");
		int counter=1;
		while(true){
			for (nbTasks = 3; nbTasks <= 10; nbTasks++) {
				ITaskSetGenerator gen = new UUniFastDiscardTaskSetGen(new DefaultTaskGenerator(), nbTasks, nbProc, U, Ue, pr, emax,0, false, true);
				for (int i = 0; i < count; i++) {
					ITask [] tasks = gen.generate();
					if(tasks != null){
						IEnergyProfile ep = EnergyProfileFactory.getInstance("default", pr, emax, 0, 0);
						ISchedulingPolicy policy = SchedulingPolicyFactory.getInstance("EH_PFP_1");

						ITaskSet taskSet = policy.createTaskSet();
						taskSet.addTasks(Arrays.asList(tasks));
						Simulation sim = new Simulation(new SimulationParameters(policy, simEnd, ep, taskSet, nbProc, isEnergyAware, isHardRealtime));
						sim.simulate();
						
						System.out.println("counter = "+counter);
						counter++;
						
						if(sim.getStatistics().isFailed()){
							ITasksetsWriter writer = TasksetsIOFactory.newTasksetsWriterInstance("txt", "file", pr);
							writer.write(taskSet);
							writer.close();
							System.out.println("finished");
							System.exit(0);
						}
					}
				}
			}
		}
	}

	/*
	public static void main(String[] args) throws Exception {
		int nbSystems = 10, pr = 35;
		long emax = 400;
		int nbTasks = 6;
		// String taskType = "simple";
		String taskType = "graph";
		try {
			nbSystems = Integer.parseInt(args[0]);
		} catch (Exception e) {
			nbSystems = 100;
		}

		// String taskSetsSource = "results/" + nbSystems + "_systems_pr_" + pr;


		  String taskSetsSource = nbSystems + taskType + "_systems_pr_" +
		 pr+".xml"; generateTaskSets(nbSystems, 2, nbTasks, pr, emax,
		  taskSetsSource, taskType);

		// XmlFileReader reader = new XmlFileReader(taskSetsSource);
		// reader.readTasksets();

		XmlFromFileWriter writer = new XmlFromFileWriter(
				SystemsGenerator.class
						.getResourceAsStream("/resources/EDF_test1"),
				new FileOutputStream("EDF_test1.xml"));
	}*/

}
