package yartiss.service.model;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.swing.BoundedRangeModel;
import javax.swing.DefaultBoundedRangeModel;

import org.jfree.data.category.DefaultCategoryDataset;

import yartiss.engine.simulation.IEnergyProfile;
import yartiss.engine.simulation.SimulationParameters;
import yartiss.engine.simulation.SimulationStatistics;
import yartiss.engine.simulation.statistics.StatisticsManager;
import yartiss.engine.simulation.system.SystemMetrics;

public class Scenario {

	private final long simulationDuration;
	private final String scenarioName;
	private long systemsCount;
	private final Map<String, Integer> nbReceivedByPolicy = new HashMap<String, Integer>();
	private final Map<String, SimulationStatistics> policiesStats = new HashMap<String, SimulationStatistics>();
	private final DefaultCategoryDataset data = new DefaultCategoryDataset();
	private final BoundedRangeModel progressModel = new DefaultBoundedRangeModel(0, 1, 0, 100000);
	private final TasksSetsTableModel setsModel = new TasksSetsTableModel();
	private final IEnergyProfile profile;
	private final float  unit = 0.05f;
	private Map<String, SimulationStatistics> [][] threedStats ;


	public Scenario(IEnergyProfile profile, long simDuaration){
		this.profile = profile;
		scenarioName =profile.getMaxEnergyLevel()+"_"+profile.getPowerRate();
		simulationDuration = simDuaration;

		int n= Math.round(1.f/unit);
		threedStats = new Map[n+1][n+1];
	}

	private void update3dStats(SimulationParameters params, SimulationStatistics newStats){
		double u  = Math.round(SystemMetrics.utilisation(params.getTaskSet()) * 1000) / 1000.0d;
		double ue = Math.round(SystemMetrics.energyUtilization(params.getTaskSet(), params.getEnergyProfile().getPowerRate()) * 1000) / 1000.0d;

		int line = (int) Math.round(u / unit);
		int row = (int) Math.round(ue / unit);

		String policyName = params.getSchedulingPolicy().getPolicyName();
		Map<String, SimulationStatistics> policies = threedStats[line][row]!=null ? threedStats[line][row] : new HashMap<String, SimulationStatistics>();
		SimulationStatistics stat = policies.containsKey(policyName) ? policies.get(policyName) : new SimulationStatistics();

		for (String criterion : StatisticsManager.getActivatedCriteria()) {
			stat.aggregate(criterion, newStats);
			stat.getCriterion(criterion).setSimulationParameters(newStats.getCriterion(criterion).getSimulationParameters());
			Object value = stat.getCriterion(criterion).getAggregatedValue();


		}

		policies.put(policyName, stat);
		threedStats[line][row] = policies;
	}


	public void updatePolicyStats(SimulationParameters params, SimulationStatistics newStats){

		update3dStats(params, newStats);

		String policyName = params.getSchedulingPolicy().getPolicyName();
		SimulationStatistics stat = policiesStats.containsKey(policyName) ? policiesStats.get(policyName) : new SimulationStatistics();
		int nbReceived = nbReceivedByPolicy.containsKey(policyName) ? nbReceivedByPolicy.get(policyName)+1 : 1 ;



		for (String criterion : StatisticsManager.getActivatedCriteria()) {
			stat.aggregate(criterion, newStats);
			stat.getCriterion(criterion).setSimulationParameters(newStats.getCriterion(criterion).getSimulationParameters());
			Object value = stat.getCriterion(criterion).getAggregatedValue();
			if(value instanceof Number)
				getData().setValue((Number)value, criterion, policyName);

		}




		policiesStats.put(policyName, stat);
		nbReceivedByPolicy.put(policyName, nbReceived);
		
		
		int newValue = 0;
		for (int val : nbReceivedByPolicy.values()) {
			newValue += val;
		}
		
		if(systemsCount==0)
			systemsCount = 10000;
		
		progressModel.setMaximum((int) systemsCount*policiesStats.size());
		progressModel.setValue(newValue);

		if(nbReceived == getSystemsCount()){
			System.out.println(policyName+" : "+stat);
		}
	}

	/**
	 * @return the scenarioName
	 */
	public String getScenarioName() {
		return scenarioName;
	}

	/**
	 * @return the data
	 */
	public DefaultCategoryDataset getData() {
		return data;
	}

	/**
	 * @return the systemsCount
	 */
	public long getSystemsCount() {
		return systemsCount;
	}

	/**
	 * @param systemsCount the systemsCount to set
	 */
	public void setSystemsCount(long systemsCount) {
		this.systemsCount = systemsCount;

	}

	/**
	 * @return the simulationDuration
	 */
	public long getSimulationDuration() {
		return simulationDuration;
	}

	/**
	 * @return the progressModel
	 */
	public BoundedRangeModel getProgressModel() {
		return progressModel;
	}

	/**
	 * @return the setsModel
	 */
	public TasksSetsTableModel getSetsModel() {
		return setsModel;
	}

	/**
	 * @return the profile
	 */
	public IEnergyProfile getProfile() {
		return profile;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String str = "Scenario "+scenarioName+" : \n";
		for (Entry<String, SimulationStatistics> policy : policiesStats.entrySet()) {
			str+= "\t"+policy.getKey()+"[";
			for (String criterion :StatisticsManager.getActivatedCriteria()) {
				str+=" "+criterion+"="+policy.getValue().getAggregatedValue(criterion)+",";
			}
			str+="]\n";
		}


		return str;
	}



	public void saveResults(String path) throws IOException{
		Writer writer = new OutputStreamWriter(new FileOutputStream(path+"/results_"+scenarioName+".data"));
		writer.write("policy\t");
		for (String criterion : StatisticsManager.getActivatedCriteria()) {
			if(StatisticsManager.getInstance(criterion).getValue() instanceof Number)
				writer.write("\""+criterion+"\"\t");
		}
		writer.write("\n");
		for (Entry<String, SimulationStatistics> policy : policiesStats.entrySet()) {
			writer.write("\""+policy.getKey()+"\"\t");
			for (String criterion : StatisticsManager.getActivatedCriteria()) {
				if(StatisticsManager.getInstance(criterion).getValue() instanceof Number)
					writer.write(policy.getValue().getAggregatedValue(criterion)+"\t");
			}
			writer.write("\n");
		}
		writer.close();

		save3DResults(path);
	}

	private void save3DResults(String path) throws IOException {

		Set<String> policiesSet = policiesStats.keySet();
		for (String criterion : StatisticsManager.getActivatedCriteria()) {
			if(StatisticsManager.getInstance(criterion).getValue() instanceof Number){
				Writer writer = new OutputStreamWriter(new FileOutputStream(path+"/results3D_"+scenarioName+"_"+criterion+".data"));
				writer.write("#\n/*-------------------------------"+criterion+"-----------------------------*/\n");
				writer.write("U\tUe\t");
				for ( String policy : policiesSet) {
					writer.write(policy+"\t");
				}
				writer.write("\n");
				for (int i = 0; i < threedStats.length; i++) {
					for (int j = 0; j < threedStats.length; j++) {
						writer.write(i*unit+"\t"+j*unit+"\t");
						for ( String policy : policiesSet) {

							Map<String, SimulationStatistics> policies = threedStats[i][j]!=null ? threedStats[i][j] : new HashMap<String, SimulationStatistics>();
							SimulationStatistics stat = policies.containsKey(policy) ? policies.get(policy) : null;

							Object value = stat!=null ? stat.getCriterion(criterion).getAggregatedValue() : 0;
							if(value instanceof Number)
								writer.write(value+"\t");
						}
						writer.write("\n");
					}
				}
				writer.close();
			}
		}
		Writer writer = new OutputStreamWriter(new FileOutputStream(path+"/results3D_"+scenarioName+".gnuplot"));
		long n = threedStats.length-Math.round(0.2/unit);
		writer.write("set hidden3d\n");
		writer.write("set dgrid3d "+n+","+n+"\n");
		writer.write("set xrange [0.2:1]\n");
		writer.write("set yrange [0.2:1]\n");
		writer.write("set xlabel \"U\" offset -3,-2\n");
		writer.write("set ylabel \"Ue\" offset 3,-2\n");

		int termId = 0;
		for (String criterion : StatisticsManager.getActivatedCriteria()) {
			if(StatisticsManager.getInstance(criterion).getValue() instanceof Number){
				writer.write("set terminal windows "+termId+"\n");
				termId++;
				writer.write("set title '"+scenarioName+"_"+criterion+"'\n");
				String splot = "splot ";
				int i=3;
				for ( String policy : policiesSet) {
					splot += "'./"+"results3D_"+scenarioName+"_"+criterion+".data' using 1:2:"+i+" with lines title '"+policy+"'";

					i++;
					if(i-3<policiesSet.size())splot += ", ";

				}
				writer.write(splot+"\n");
			}
		}
		writer.close();

	}


}
